
/*

   muscan - scanner driver for Mustek Paragon 6000CX scanner
   Copyright (C) 1995 Torsten Eichner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   email:       eichner@rhrk.uni-kl.de
   papermail:   Torsten Eichner, Kampshardtweg 1, D-57587 Birken-Honigsessen, Germany

 */

/*
   Start of file: muscan.c
 */

#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef LINUX
#include </usr/src/linux/drivers/scsi/sg.h>
#endif

#ifdef FREEBSD
#include <sys/scsiio.h>
#include <scsi.h>
#endif

#define MAXCOLS 3000
#define MAXROWS 43000
#define MAXHIST 500
#define PBM_WHITE 0
#define PBM_BLACK 1

typedef unsigned char bit;

static void usage (char *);
static void parse_commandline (int, char *[]);
static void open_scanner (void);
static void close_scanner (void);
static void write_print_image (unsigned char *, int, int);
static void write_image (unsigned char *, int, int);
static void signal_handler (int);
static int scsi_cmd (unsigned char *, int, unsigned char *, int);

static void scan_image_6000cx (int);
static void scan_image_6000cz (int);
static int image_status_6000cz (int *, int *);
static void scan_image_6000sp (int);

static char *scanner_device = DEVICE;
static int scanner_fd = -1;
static int scanner_index = 0;

static char *outputfile = "-";
static int scan_running = 0;

static float x_size = 8.3;
static float y_size = 11.7;
static float x_pos = 0;
static float y_pos = 0;

static int resolution = 100;
static int color = 0;
static int brightness = 12;
static int contrast = 12;
static int grain = 0;
static int speed = 1;
static int backtracking = -1;
static int outputformat = 0;

static int o_item = -1;
static int oo_item = -1;
static char *cb;
static int cmode = 2;
static int item, bitsperitem;
static char *line;
static int bytecnt;

static struct
  {
    int bytecnt;
    char *line;
  }
hist[MAXHIST];

static int maxhist = MAXHIST;
static int histidx = 0;
static int lncnt = 0;
static int maxline = 13 * 300;

static int mem_size = 4096;

static int debug = 0;

static struct
  {
    char *name;
    void (*scan) ();
    int param;
  }
scanner_type[] =
{
  {
    "MFS-06000CX", scan_image_6000cx, 0
  }
  ,
  {
    "MFS-06000CZ", scan_image_6000cz, 0
  }
  ,
  {
    "MSF-06000CX", scan_image_6000cx, 0
  }
  ,
  {
    "MSF-06000CZ", scan_image_6000cz, 0
  }
  ,
  {
    "MSF-08000CZ", scan_image_6000cz, 1
  }
  ,
  {
    "MFS-08000CZ", scan_image_6000cz, 1
  }
  ,
  {
    "MSF-12000CZ", scan_image_6000cz, 2
  }
  ,
  {
    "MFS-12000CZ", scan_image_6000cz, 2
  }
  ,
  {
    "MSF-06000SP", scan_image_6000sp, 0
  }
  ,
  {
    "MSF-08000SP", scan_image_6000sp, 1
  }
  ,
  {
    "MSF-12000SP", scan_image_6000sp, 2
  }
  ,
  {
    "MFS-06000SP", scan_image_6000sp, 0
  }
  ,
  {
    "MFS-08000SP", scan_image_6000sp, 1
  }
  ,
  {
    "MFS-12000SP", scan_image_6000sp, 2
  }
  ,
  {
    NULL
  }
};

#define ERROR(string...) { fprintf(stderr, "ERROR[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string); exit(-1); }
#define TRACE(lvl, string...) { if (debug >= lvl) { fprintf(stderr, "TRACE[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string);fflush(stderr); } }

/*******************************************
 * Print help                              *
 *******************************************/
static void usage
  (
    char *text			/* I: NULL or erroneous argument */
)
{
  printf ("muscan V%s\n", VERSION);

  printf ("Syntax: muscan [-D] [(+|-)b] [-c] [-d] [-f outputfile] [-h] [-l pos] [-r resolution] [-u pos] [-v] [-x length] [-y length] [-?]\n");
  printf ("Options:\n");
  printf (" -D      : generate Deskjet output (default is PPM)\n");
  printf (" -a      : use lineart-mode (default is greyscale)\n");
  printf (" -b      : turn off backtracking (default for SP models)\n");
  printf (" +b      : turn on backtracking (default for CX and CZ models)\n");
  printf (" -c      : use colormode (default is greyscale)\n");
  printf (" -d      : turn on debugging\n");
  printf (" -f file : copy to file instead of stdout\n");
  printf (" -h      : print this help\n");
  printf (" -l pos  : set x position of image (default 0\")\n");
  printf (" -m      : use halftone-mode (default is greyscale)\n");
  printf (" -p block: set various parameters (block=brightness:contrast:grain:speed, default 12:12:0:1)\n");
  printf (" -r res  : set resolution to res dpi (default 100)\n");
  printf (" -u pos  : set y position of image (default 0\")\n");
  printf (" -v      : show version\n");
  printf (" -x len  : set x size in inch (default 8.3\" = DIN A4)\n");
  printf (" -y len  : set y size in inch (default 11.7\" = DIN A4)\n");
  printf (" -?      : print this help\n");

  if (text)
    printf ("Erroneous argument: <%s>\n", text);

  if (text)
    exit (-1);
  else
    exit (0);
}

/*******************************************
 * Parse commandline                       *
 *******************************************/
static void parse_commandline
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  int i;

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
	{
	case 'D':		/* Deskjet output */
	  outputformat = 1;
	  break;

	case 'a':		/* line-art */
	  color = 3;
	  break;

	case 'b':		/* backtracking */
	  backtracking = 0;
	  break;

	case 'c':		/* color mode */
	  color = 1;
	  break;

	case 'd':		/* debug mode */
	  debug++;
	  break;

	case 'f':		/* outputfile */
	  i++;
	  if (i < argc)
	    outputfile = argv[i];
	  else
	    usage (argv[i - 1]);
	  break;

	case 'h':		/* help */
	  usage (NULL);
	  break;

	case 'l':		/* x-pos */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &x_pos) != 1 || x_pos < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'm':		/* halftone mode */
	  color = 2;
	  break;

	case 'p':		/* various parameters */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%d:%d:%d:%d", &brightness, &contrast, &grain, &speed) != 4 || brightness < 0 || contrast < 0 || grain < 0 || speed < 0 || brightness > 24 || contrast > 24 || grain > 5 || speed > 4)
	    usage (argv[i - 1]);
	  break;

	case 'r':		/* resolution */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%d", &resolution) != 1 || resolution < 1 || resolution > 1200)
	    usage (argv[i - 1]);
	  break;

	case 'u':		/* y-pos */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &y_pos) != 1 || y_pos < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'v':		/* version */
	  printf ("muscan V%s - compiled by %s on %s\n", VERSION, USER, HOST);
	  printf ("Copyright (C) 1995 by Torsten Eichner\n");
	  printf ("E-mail: eichner@rhrk.uni-kl.de\n");
	  printf ("This program is distributed under the GNU GPL - please refer to it for further information.\n");
	  break;

	case 'x':		/* x-size */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &x_size) != 1 || x_size < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'y':		/* y-size */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &y_size) != 1 || y_size < 0.0)
	    usage (argv[i - 1]);
	  break;

	case '?':		/* help */
	  usage (NULL);
	  break;

	default:
	  usage (argv[i]);
	  break;
	}
    else if (argv[i][0] == '+')
      switch (argv[i][1])
	{
	case 'b':		/* backtracking */
	  backtracking = 1;
	  break;

	default:
	  usage (argv[i]);
	  break;
	}
    else
      usage (argv[i]);

  TRACE (1, "scanning %2.2f\"x%2.2f\" %s-image to <%s> at pos %2.2f\":%2.2f\" with %d dpi\n", x_size, y_size, color ? "color" : "greyscale", outputfile, x_pos, y_pos, resolution);
  TRACE (1, "pblock %d:%d:%d:%d\n", brightness, contrast, grain, speed);
}

/*******************************************
 * Signal handler                          *
 *******************************************/
static void signal_handler
  (
    int signal			/* I: received signal */
)
{
  if (scanner_fd != -1)
    {
      char stop[] =
      {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};

      if (scan_running)
	scsi_cmd (stop, 6, NULL, 0);
      scan_running = 0;
      close_scanner ();
    }
  exit (1);
}

#ifdef LINUX
/*******************************************
 * Issue a SCSI command                    *
 *******************************************/
static int scsi_cmd
  (
    unsigned char *src,		/* I: source data */
    int src_size,		/* I: length of source data */
    unsigned char *dest,	/* O: result */
    int dest_size		/* I: expected size of result */
)
{
#ifndef SG_BIG_BUFF
  unsigned char scsi_buffer[4096 + sizeof (struct sg_header)];
#else
  unsigned char scsi_buffer[SG_BIG_BUFF];
#endif
  struct sg_header hdr;
  int i, count;

  TRACE (2, "issue SCSI command 0x%02x (expected replylen: %d+%d=%d)\n", src[0], dest_size, sizeof (struct sg_header), dest_size + sizeof (struct sg_header));

#ifdef SG_BIG_BUF
  if (dest_size > SG_BIG_BUF - sizeof (struct sg_header))
      ERROR ("buffer overflow: expected %d bytes, buffer is %d bytes\n", count, SG_BIG_BUF - sizeof (struct sg_header));
#endif

  hdr.pack_len = src_size + sizeof (struct sg_header);
  hdr.reply_len = dest_size + sizeof (struct sg_header);

  for (i = 0; i < sizeof (struct sg_header); i++)
    scsi_buffer[i] = ((char *) &hdr)[i];

  for (; i < hdr.pack_len; i++)
    scsi_buffer[i] = src[i - sizeof (struct sg_header)];

  if (write (scanner_fd, scsi_buffer, hdr.pack_len) != hdr.pack_len)
    ERROR ("error on write: %s\n", strerror (errno));

  if ((count = read (scanner_fd, scsi_buffer, dest_size + sizeof (struct sg_header))) < 0)
      ERROR ("error on read: %s\n", strerror (errno));

  for (i = 0; i < sizeof (struct sg_header); i++)
    ((char *) &hdr)[i] = scsi_buffer[i];

  if (hdr.result != 0)
    ERROR ("SCSI error (%02x): (cmd %02x) sense 0x%02x 0x%02x 0x%02x...\n", hdr.result, src[0], hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2]);

  TRACE (2, "SCSI: read %4d packlen %4d replylen %d result 0x%02x\n", count, hdr.pack_len, hdr.reply_len, hdr.result);
  TRACE (2, "SCSI: sense %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2], hdr.sense_buffer[3], hdr.sense_buffer[4], hdr.sense_buffer[5], hdr.sense_buffer[6], hdr.sense_buffer[7], hdr.sense_buffer[8], hdr.sense_buffer[9], hdr.sense_buffer[10], hdr.sense_buffer[11], hdr.sense_buffer[12], hdr.sense_buffer[13], hdr.sense_buffer[14], hdr.sense_buffer[15]);

  TRACE (3, "SCSI data (BEGIN)\n");

  for (; i < count; i++)
    {
      dest[i - sizeof (struct sg_header)] = scsi_buffer[i];
      if ((src[0] != 0x08 && debug > 2) || (debug > 3))
	fprintf (stderr, "%4x:%02x(%c)%s", i - sizeof (struct sg_header), scsi_buffer[i], scsi_buffer[i] >= 32 && scsi_buffer[i] < 127 ? scsi_buffer[i] : ' ', (i - sizeof (struct sg_header) + 1) % 8 == 0 ? "\n" : " ");
    }

  TRACE (3, "SCSI data (END)\n");

  return count - sizeof (struct sg_header);
}
#endif

#ifdef FREEBSD
/*******************************************
 * Issue a SCSI command                    *
 *******************************************/
static int scsi_cmd
  (
    unsigned char *src,		/* I: source data */
    int src_size,		/* I: length of source data */
    unsigned char *dest,	/* O: result */
    int dest_size		/* I: expected size of result */
)
{
  char cmd[50], *p1;
  int cmdlen;
  static scsireq_t req;

  p1 = cmd;
  cmdlen = (*src == 0x55) ? 10 : 6;

  scsireq_reset (&req);

  while (src_size-- && cmdlen--)
    {
      sprintf (p1, "%02x ", (int) *src);
      p1 += strlen (p1);
      src++;
    }
  ++src_size;
  if (src_size)
    scsireq_build (&req, src_size, src, SCCMD_WRITE, cmd);
  else if (dest_size)
    scsireq_build (&req, dest_size, dest, SCCMD_READ, cmd);
  else
    scsireq_build (&req, 0, NULL, 0, cmd);
/*
   req.timeout = 200000;
 */
  req.timeout = 200000;

  if (scsireq_enter (scanner_fd, &req) == -1)
    perror ("scsi_cmd");
  return 0;
}
#endif

/*******************************************
 * Open scanner                            *
 *******************************************/
static void open_scanner
  (
    void
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char result[256];
#ifdef SG_BIG_BUFF
  union
    {
      unsigned char tmp[SG_BIG_BUFF + sizeof (struct sg_header)];
      struct sg_header hdr;
    }
  sgb;
#endif
  int i;
  int valid_size;

  TRACE (1, "opening scanner <%s>\n", scanner_device);

#ifdef LINUX
  if ((scanner_fd = open (scanner_device, O_RDWR | O_EXCL)) < 0)
    ERROR ("cannot open scanner <%s> due to: %s\n", scanner_device, strerror (errno));
#endif

#ifdef FREEBSD
  if ((scanner_fd = scsi_open (scanner_device, O_RDWR | O_EXCL)) < 0)
    ERROR ("cannot open scanner <%s> due to: %s\n", scanner_device, strerror (errno));
#endif

  scsi_cmd (inquiry, 6, result, 256);

  if (result[0] != 6 || strncmp (result + 8, "MUSTEK", 6))
    ERROR ("no MUSTEK scanner found\n");

  for (i = 0; scanner_type[i].name != NULL; i++)
    if (strncmp (scanner_type[i].name, result + 16, strlen (scanner_type[i].name)) == 0)
      break;

  scanner_index = i;

  if (scanner_type[scanner_index].name == NULL)
    ERROR ("unknown scanner model %s\n", result + 16);

  TRACE (1, "scanner type %s (%d)\n", scanner_type[scanner_index].name, scanner_index);

#ifdef SG_BIG_BUFF
  /* try to determine maximum buffer size */
  for (valid_size = 4096; valid_size < SG_BIG_BUFF; valid_size += 1024)
    {
      sgb.hdr.pack_len = 6 + sizeof (struct sg_header);
      sgb.hdr.reply_len = valid_size;

      for (i = 0; i < 6; i++)
	sgb.tmp[i + sizeof (struct sg_header)] = inquiry[i];

      if (write (scanner_fd, &sgb, 6 + sizeof (struct sg_header)) < 0)
	  break;

      read (scanner_fd, &sgb, sgb.hdr.reply_len + sizeof (struct sg_header));

      mem_size = valid_size;
    }

  TRACE (1, "new mem_size is %d\n", mem_size);
#endif
}

/*******************************************
 * Close scanner                           *
 *******************************************/
static void close_scanner
  (
    void
)
{
  TRACE (1, "closing scanner <%s>\n", scanner_device);

  close (scanner_fd);
  scanner_fd = -1;
}

/*******************************************
 * The following are routines from g3tolj  *
 *******************************************/
static void putinit
  (
    FILE * fd
)
{
  int i;

  for (i = 0; i < maxhist; i++)
    {
      hist[i].line = (char *) malloc (2 * resolution * 10);
    }
  histidx = 0;
  lncnt = 0;
  line = hist[histidx].line;

  /* Printer reset. */
  fprintf (fd, "\033E");
  fprintf (fd, "\033*rB");

  fprintf (fd, "\033*t%dR", resolution);	/* Set raster graphics resolution */
  /* move to top left of page & set current position */
  fprintf (fd, "\033&l2A");	/* A4 size */
  fprintf (fd, "\033&l0L");	/* Don't skip perforation */
  fprintf (fd, "\033&l0E");	/* Top margin 0 */
  fprintf (fd, "\033*p0x0Y");
  fprintf (fd, "\033*r1A");	/* Start raster graphics, relative adressing */
  /* printf("\033*b2M"); */
  fprintf (fd, "\033*b%dM", cmode);	/* set compression mode, 1=rll */
}

static void putitem
  (
    void
)
{
  if (cmode == 1)
    {
      if (item == o_item && ((unsigned char) (line[bytecnt - 2]) < 255))
	{
	  ++line[bytecnt - 2];
	  bitsperitem = item = 0;
	  return;
	}
      line[bytecnt++] = 0;	/* repeat count of one */
      line[bytecnt++] = o_item = item;
      bitsperitem = 0;
      item = 0;
      return;
    }
  if (cmode == 2)
    {
      /* if running repeat, continue if same char again */
      if (item == o_item && cb != NULL && *cb < 1 && *cb > -127)
	{
	  --*cb;
	  bitsperitem = item = 0;
	  return;
	}
      /* if running literal see last two same as next and switch to repeat */
      if (cb != NULL && *cb > 0 && o_item == oo_item && o_item == item)
	{
	  *cb -= 2;
	  cb = &line[bytecnt - 2];
	  *cb = -2;
	  bitsperitem = item = 0;
	  return;
	}
      /* if running literal add next byte to string */
      if (cb != NULL && *cb >= 0 && *cb < 127)
	{
	  ++*cb;
	  oo_item = o_item;
	  line[bytecnt++] = o_item = item;
	  bitsperitem = item = 0;
	  return;
	}
      /* start new literal */
      cb = &line[bytecnt];
      line[bytecnt++] = 0;	/* repeat count of one */
      line[bytecnt++] = o_item = item;
      oo_item = -1;
      bitsperitem = 0;
      item = 0;
      return;
    }
  line[bytecnt++] = item;
  bitsperitem = 0;
  item = 0;
}

static void putbit
  (
    bit b
)
{
  if (b == PBM_BLACK)
    item += 1 << (7 - bitsperitem);
  bitsperitem++;
  if (bitsperitem == 8)
    {
      putitem ();
    }
}

static void putrest
  (
    FILE * fd
)
{
  /* end raster graphics */
  fprintf (fd, "\033*rB");

  /* Printer reset. */
  fprintf (fd, "\033E");
}

static void newpage
  (
    FILE * fd
)
{
  int i;

  putrest (fd);
  putinit (fd);
  lncnt = 0;
  for (i = 0; i < maxhist; i++)
    {
      fprintf (fd, "\033*b%dW", hist[histidx].bytecnt);
      fwrite (hist[histidx].line, sizeof (char), hist[histidx].bytecnt, fd);

      histidx++;
      if (histidx >= maxhist)
	histidx = 0;
      lncnt++;
    }
}

static void lj_writerow
  (
    FILE * fd,
    bit * writerow,
    int wcols
)
{
  int i;

  oo_item = o_item = -1;
  cb = NULL;
  item = 0;
  bytecnt = 0;
  bitsperitem = 0;
  for (i = 0; i < wcols; i++)
    putbit (writerow[i]);
  if (bitsperitem > 0)
    putitem ();
  fprintf (fd, "\033*b%dW", bytecnt);
  fwrite (line, sizeof (char), bytecnt, fd);

  hist[histidx].bytecnt = bytecnt;
  histidx++;
  if (histidx >= maxhist)
    histidx = 0;
  line = hist[histidx].line;
  lncnt++;
  if (lncnt > maxline)
    newpage (fd);
};
/*******************************************
 * End of "imported" section               *
 *******************************************/

/*******************************************
 * Write print image                       *
 *******************************************/
static void write_print_image
  (
    unsigned char *image,	/* I: input image */
    int x_size,			/* I: pixel per line */
    int y_size			/* I: lines */
)
{
  FILE *stream;
  int x, y;
  bit linebuf[5000];
  bit *linebufp;
  int p;

  TRACE (1, "writing image to <%s>\n", outputfile);

  if (strcmp (outputfile, "-"))
    {
      if ((stream = fopen (outputfile, "w")) == NULL)
	ERROR ("cannot open <%s>: %s\n", outputfile, strerror (errno));
    }
  else
    stream = stdout;

  switch (color)
    {
    case 0:			/* grey */
    case 1:			/* color */
      ERROR ("Format not supported for deskjet output");
      break;

    case 2:			/* b/w lineart */
    case 3:			/* b/w halftone */
      TRACE (1, "deskjet image  <%4d x %4d>\n", x_size, y_size);
      putinit (stream);
      for (y = 0; y < (11.7 * resolution - y_size) / 2; y++)
	{
	  linebuf[0] = PBM_WHITE;
	  lj_writerow (stream, linebuf, 1);
	}
      for (y = 0; y < y_size; y++)
	{
	  linebufp = linebuf;
	  for (x = 0; x < (8.3 * resolution - x_size) / 2; x++)
	    *linebufp++ = PBM_WHITE;
	  for (x = 0; x < x_size / 8; x++)
	    for (p = 0; p < 8; p++)
	      *linebufp++ = (image[y * x_size + x] & (1 << (7 - p))) == 0;
	  lj_writerow (stream, linebuf, linebufp - linebuf);
	}
      putrest (stream);
      break;
    }
  if (strcmp (outputfile, "-"))
    fclose (stream);
}

/*******************************************
 * Write image                             *
 *******************************************/
static void write_image
  (
    unsigned char *image,	/* I: input image */
    int x_size,			/* I: pixel per line */
    int y_size			/* I: lines */
)
{
  FILE *stream;
  int x, y;

  TRACE (1, "writing %dx%d image to <%s>\n", x_size, y_size, outputfile);

  if (strcmp (outputfile, "-"))
    {
      if ((stream = fopen (outputfile, "w")) == NULL)
	ERROR ("cannot open <%s>: %s\n", outputfile, strerror (errno));
    }
  else
    stream = stdout;

  switch (color)
    {
    case 0:			/* grey */
      fprintf (stream, "P5\n#CREATOR: muscan Version %s\n%d %d\n255\n", VERSION, x_size, y_size);

      fwrite (image, 1, x_size * y_size, stream);
      break;

    case 1:			/* color */
      fprintf (stream, "P6\n#CREATOR: muscan Version %s\n%d %d\n255\n", VERSION, x_size, y_size);

      for (y = 0; y < y_size; y++)
	for (x = 0; x < x_size; x++)
	  {
	    unsigned char pixel[3];
	    int ofs;

	    ofs = x + y * x_size;

	    pixel[0] = image[ofs];
	    pixel[1] = image[ofs + x_size * y_size];
	    pixel[2] = image[ofs + x_size * y_size * 2];

	    fwrite (pixel, 3, 1, stream);
	  }
      break;

    case 2:			/* b/w lineart */
    case 3:			/* b/w halftone */
      fprintf (stream, "P4\n#CREATOR: muscan Version %s\n%d %d\n255\n", VERSION, (x_size / 8) * 8, y_size);

      for (x = 0; x < x_size * y_size; x++)
	image[x] = ~image[x];

      for (y = 0; y < y_size; y++)
	fwrite (image + y * x_size, 1, x_size / 8, stream);
      break;
    }

  if (strcmp (outputfile, "-"))
    fclose (stream);
}

/*******************************************
 * Scan image (6000CX)                     *
 *******************************************/
static void scan_image_6000cx
  (
    int param			/* I: scanner type, nc */
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char testunit[] =
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char area[] =
  {0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0, 0, 0, 0, 0, 0, 0, 0};
  unsigned char mode[] =
  {0x15, 0x00, 0x00, 0x00, 0x0b, 0x00, 0x83, 0x00, 0x0c, 0x0c, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00};
  unsigned char lookup[] =
  {0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char start[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char stop[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char status[] =
  {0x0f, 0x00, 0x00, 0x00, 0x06, 0x00};
  unsigned char backtrack[] =
  {0x10, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char readline[] =
  {0x08, 0x00, 0x00, 0x00, 0x01, 0x00};
#ifndef SG_BIG_BUFF
  unsigned char image[4096];
#else
  unsigned char image[SG_BIG_BUFF];
#endif
  unsigned char state[6];
  int run, i, line, maxline, bpl, maxbpl;
  int x, y;
  unsigned char *slice = NULL;
  int ll;

#define MAXRETRY 8
#define BIGRETRY 3

  TRACE (1, "scan image\n");

  maxbpl = resolution * x_size;
  maxline = resolution * y_size;

  if (resolution > 300 || resolution < 3)
    ERROR ("need resolution in the range of 3..300\n");

  if (y_size + y_pos > 14.0)
    ERROR ("illegal size/length: would run over border by %2.2f\"\n", y_size + y_pos - 14.0);

  if (x_size + x_pos > 9.0)
    ERROR ("illegal size/width: would run over border by %2.2f\"\n", x_size + x_pos - 9.0);

  if (maxbpl >= 4095)
    ERROR ("line overflow (%d, allowed 4095) - use lower resolution or x-size\n", maxbpl);

  if ((slice = malloc (maxbpl * maxline * (color == 1 ? 3 : 1))) == NULL)
    ERROR ("cannot malloc %d bytes\n", maxbpl * maxline * (color == 1 ? 3 : 1));

  memset (slice, 0, maxbpl * maxline * (color == 1 ? 3 : 1));

  for (run = 0; run < (color == 1 ? 3 : 1); run++)
    {
      int busy, lib, retry, bigretry;
      unsigned char mask;

      bigretry = 0;
      retry = 0;

    scan_retry:
      scsi_cmd (inquiry, 6, image, 256);

      scsi_cmd (testunit, 6, NULL, 0);

      area[6] = (color == 3 ? 0 : 1);
      area[7] = ((unsigned int) (x_pos * 8)) & 0xff;
      area[8] = (((unsigned int) (x_pos * 8)) & 0xff00) >> 8;
      area[9] = ((unsigned int) (y_pos * 8)) & 0xff;
      area[10] = (((unsigned int) (y_pos * 8)) & 0xff00) >> 8;

      area[11] = ((unsigned int) ((x_pos + x_size) * 8)) & 0xff;
      area[12] = (((unsigned int) ((x_pos + x_size) * 8)) & 0xff00) >> 8;
      area[13] = ((unsigned int) ((y_pos + y_size) * 8)) & 0xff;
      area[14] = (((unsigned int) ((y_pos + y_size) * 8)) & 0xff00) >> 8;

      scsi_cmd (area, 15, NULL, 0);

      mode[7] = (resolution / 3) & 0xff;
      mode[8] = brightness & 0xff;
      mode[9] = contrast & 0xff;
      mode[10] = grain & 0xff;
      mode[11] = speed & 0xff;
      mode[14] = ((unsigned int) ((x_size + x_pos) * 8)) & 0xff;
      mode[15] = (((unsigned int) ((x_size + x_pos) * 8)) & 0xff00) >> 8;
      scsi_cmd (mode, 16, NULL, 0);

      scsi_cmd (lookup, 10, NULL, 0);

      mask = (color == 1) ? (run == 0 ? 0x08 : (run == 1 ? 0x10 : 0x18)) : 0;

      if (color == 0 || color == 1)
	start[4] = 0x41 | mask;
      else
	start[4] = 0x01 | mask;
      scsi_cmd (start, 6, NULL, 0);
      scan_running = 1;

      line = 0;

      retry = 0;

      do
	{
	  scsi_cmd (status, 6, state, 6);
	  busy = state[0];
	  bpl = state[1] + state[2] * 256;
	  lib = state[3] + state[4] * 256 + state[5] * 65536;

	  TRACE (2, "rdy %d bpl(%d) %4d lib(%d) %8d\n", !busy, maxbpl, bpl, lib, maxline);

	  retry++;
	  if (busy)
	    sleep (retry);

	  if (retry > MAXRETRY)
	    break;
	}
      while (busy);

      if (bpl > 4095 || busy)
	{
	  if (bigretry < BIGRETRY)
	    {
	      bigretry++;
	      scsi_cmd (stop, 6, NULL, 0);
	      scan_running = 0;

	      sleep (2);

	      TRACE (1, "bigretry %d\n", bigretry);

	      goto scan_retry;
	    }
	}

      backtrack[4] = (backtracking ? 0x82 : 0x80);
      scsi_cmd (backtrack, 6, NULL, 0);

      if (retry <= MAXRETRY)
	do
	  {
#ifdef SG_BIG_BUFF
	    ll = (mem_size - sizeof (struct sg_header)) / bpl;
	    if (ll > maxline - line)
	      ll = maxline - line;
	    if (ll < 1)
	      ll = 1;
#else
	    ll = 1;
#endif

	    readline[2] = ((ll & 0xff0000) >> 16) & 0xff;
	    readline[3] = ((ll & 0x00ff00) >> 8) & 0xff;
	    readline[4] = ((ll & 0x0000ff)) & 0xff;

	    scsi_cmd (readline, 6, image, bpl * ll);

	    for (y = 0; y < ll; y++)
	      for (x = 0; x < bpl && x < maxbpl; x++)
		slice[x + line * maxbpl + y * maxbpl + run * maxbpl * maxline] = image[x + y * bpl];

	    line += ll;
	  }
	while (line < maxline);

      TRACE (1, "lines processed %d, bpl processed %d, maxbpl %d, maxline %d\n", line, bpl, maxbpl, maxline);

      /* seems to avoid lockups, chel */
      scsi_cmd (lookup, 10, NULL, 0);

      scsi_cmd (stop, 6, NULL, 0);
      scan_running = 0;

      sleep ((int) (y_pos + y_size) + 3);

      for (i = 1; i <= MAXRETRY; i++)
	{
	  scsi_cmd (status, 6, state, 6);
	  if (!state[0])
	    break;
	  sleep (i);
	}

      scsi_cmd (stop, 6, NULL, 0);
      scan_running = 0;

      if (retry > MAXRETRY)
	ERROR ("scanner timed out\n");

      if (bpl > 4095)
	ERROR ("input line too long (%d, allowed 4095) - use lower resolution or x-size\n", bpl);
    }

  if (outputformat == 0)
    write_image (slice, maxbpl, maxline);
  else
    write_print_image (slice, maxbpl, maxline);

  free (slice);
}

/*******************************************
 * Get image status (6000CZ)               *
 *******************************************/
static int image_status_6000cz
  (
    int *bpl,			/* O: bits per line */
    int *lib			/* O: lines in buffer */
)
{
  unsigned char request_sense[] =
  {3, 0, 0, 0, 4, 0};
  unsigned char status[] =
  {0x0f, 0x00, 0x00, 0x00, 0x06, 0x00};
  unsigned char sense[16];
  unsigned char state[6];
  static int lastlib = 0, lastbpl = 0;

  scsi_cmd (request_sense, 6, sense, 16);

  scsi_cmd (status, 6, state, 6);
  *bpl = state[1] + (state[2] << 8);
  *lib = state[3] + (state[4] << 8) + (state[5] << 16);

  if (*lib > (1 << 23))
    *lib = -((~*lib) & ((1 << 23) - 1)) - 1;

  if (*bpl != lastbpl)
    TRACE (1, "image_status:bpl changed from %d to %d\n", lastbpl, *bpl);
  lastbpl = *bpl;

  if (lastlib != *lib)
    TRACE (1, "image_status:Lines-in-Buffer changed from %d to %d\n", lastlib, *lib);
  lastlib = *lib;

  TRACE (1, "image_status:%s\n", state[0] ? "NOT READY" : "Ready");

  return state[0];
}

/*******************************************
 * Scan image (6000CZ)                     *
 *******************************************/
static void scan_image_6000cz
  (
    int param			/* I: scanner type, nc */
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char testunit[] =
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char area[] =
  {0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0, 0, 0, 0, 0, 0, 0, 0};
  unsigned char mode[] =
  {0x15, 0x00, 0x00, 0x00, 0x0b, 0x00, 0x83, 0x00, 0x0c, 0x0c, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00};
  unsigned char lookup[] =
  {0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char start[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char stop[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char status[] =
  {0x0f, 0x00, 0x00, 0x00, 0x06, 0x00};
  unsigned char backtrack[] =
  {0x10, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char readline[] =
  {0x08, 0x00, 0x00, 0x00, 0x01, 0x00};
#ifndef SG_BIG_BUFF
  unsigned char image[4096];
#else
  unsigned char image[SG_BIG_BUFF];
#endif
  unsigned char state[6];
  int run, i, line, maxline, bpl = 0, lib = 0, maxbpl, dlib, dbpl, area_retries;
  int x, y;
  unsigned char *slice = NULL;
  int ll;
  int allowed_res[3] =
  {300, 400, 600};
  int divisor[3] =
  {3, 4, 6};

#define MAXRETRY 8
#define BIGRETRY 3

  TRACE (1, "scan image\n");

  maxbpl = resolution * x_size;
  maxline = resolution * y_size;

  if (resolution > allowed_res[param] || resolution < 3)
    ERROR ("need resolution in the range of 3..%d\n", allowed_res[param]);

  if (y_size + y_pos > 14.0)
    ERROR ("illegal size/length: would run over border by %2.2f\"\n", y_size + y_pos - 14.0);

  if (x_size + x_pos > 9.0)
    ERROR ("illegal size/width: would run over border by %2.2f\"\n", x_size + x_pos - 9.0);

  if (maxbpl >= 4095)
    ERROR ("line overflow (%d, allowed 4095) - use lower resolution or x-size\n", maxbpl);

  if ((slice = malloc (maxbpl * maxline * (color == 1 ? 3 : 1))) == NULL)
    ERROR ("cannot malloc %d bytes\n", maxbpl * maxline * (color == 1 ? 3 : 1));

  memset (slice, 0, maxbpl * maxline * (color == 1 ? 3 : 1));

  for (run = 0; run < (color == 1 ? 3 : 1); run++)
    {
      int busy, retry, bigretry;
      unsigned char mask;

      bigretry = 0;
      retry = 0;

    scan_retry:
      scsi_cmd (inquiry, 6, image, 256);

      scsi_cmd (testunit, 6, NULL, 0);

      area[6] = (color == 3 ? 0 : 1);
      area[7] = ((unsigned int) (x_pos * 8)) & 0xff;
      area[8] = (((unsigned int) (x_pos * 8)) & 0xff00) >> 8;
      area[9] = ((unsigned int) (y_pos * 8)) & 0xff;
      area[10] = (((unsigned int) (y_pos * 8)) & 0xff00) >> 8;

      area[11] = ((unsigned int) ((x_pos + x_size) * 8)) & 0xff;
      area[12] = (((unsigned int) ((x_pos + x_size) * 8)) & 0xff00) >> 8;
      area[13] = ((unsigned int) ((y_pos + y_size) * 8)) & 0xff;
      area[14] = (((unsigned int) ((y_pos + y_size) * 8)) & 0xff00) >> 8;

      scsi_cmd (area, 15, NULL, 0);

      area_retries = 0;
      while (image_status_6000cz (&dbpl, &dlib))
	{
	  area_retries++;
	  if (area_retries > 10)
	    ERROR ("too many area retries(%d)\n", area_retries);

	  scsi_cmd (area, 15, NULL, 0);
	}

      mode[7] = (resolution / divisor[param]) & 0xff;
      mode[8] = brightness & 0xff;
      mode[9] = contrast & 0xff;
      mode[10] = grain & 0xff;
      mode[11] = speed & 0xff;
      mode[14] = ((unsigned int) ((x_size + x_pos) * 8)) & 0xff;
      mode[15] = (((unsigned int) ((x_size + x_pos) * 8)) & 0xff00) >> 8;
      scsi_cmd (mode, 16, NULL, 0);

      scsi_cmd (lookup, 10, NULL, 0);

      mask = (color == 1) ? (run == 0 ? 0x08 : (run == 1 ? 0x10 : 0x18)) : 0;

      if (color == 0 || color == 1)
	start[4] = 0x41 | mask;
      else
	start[4] = 0x01 | mask;
      scsi_cmd (start, 6, NULL, 0);
      scan_running = 1;

      line = 0;

      retry = 0;

      do
	{
	  busy = image_status_6000cz (&bpl, &lib);

	  TRACE (2, "rdy %d bpl(%d) %4d lib(%d) %8d\n", !busy, maxbpl, bpl, lib, maxline);

	  retry++;
	  if (busy)
	    sleep (retry);

	  if (retry > MAXRETRY)
	    break;
	}
      while (busy);

      if (bpl > 4095 || busy)
	{
	  if (bigretry < BIGRETRY)
	    {
	      bigretry++;
	      scsi_cmd (stop, 6, NULL, 0);
	      scan_running = 0;

	      sleep (2);

	      TRACE (1, "bigretry %d\n", bigretry);

	      goto scan_retry;
	    }
	}

      backtrack[4] = (backtracking ? 0x82 : 0x80);
      scsi_cmd (backtrack, 6, NULL, 0);

      if (retry <= MAXRETRY)
	do
	  {
#ifdef SG_BIG_BUFF
	    ll = (mem_size - sizeof (struct sg_header)) / bpl;
	    if (ll > maxline - line)
	      ll = maxline - line;
	    if (ll < 1)
	      ll = 1;
#else
	    ll = 1;
#endif

	    readline[2] = ((ll & 0xff0000) >> 16) & 0xff;
	    readline[3] = ((ll & 0x00ff00) >> 8) & 0xff;
	    readline[4] = ((ll & 0x0000ff)) & 0xff;

	    scsi_cmd (readline, 6, image, bpl * ll);

	    for (y = 0; y < ll; y++)
	      for (x = 0; x < bpl && x < maxbpl; x++)
		slice[x + line * maxbpl + y * maxbpl + run * maxbpl * maxline] = image[x + y * bpl];

	    line += ll;
	  }
	while (line < maxline);

      TRACE (1, "lines processed %d, bpl processed %d, maxbpl %d, maxline %d\n", line, bpl, maxbpl, maxline);

      scsi_cmd (stop, 6, NULL, 0);
      scan_running = 0;

      for (i = 1; i <= MAXRETRY; i++)
	{
	  scsi_cmd (status, 6, state, 6);
	  if (!state[0])
	    break;
	  sleep (i);
	}

      scsi_cmd (stop, 6, NULL, 0);
      scan_running = 0;

      if (retry > MAXRETRY)
	ERROR ("scanner timed out\n");

      if (bpl > 4095)
	ERROR ("input line too long (%d, allowed 4095) - use lower resolution or x-size\n", bpl);
    }

  if (outputformat == 0)
    write_image (slice, maxbpl, maxline);
  else
    write_print_image (slice, maxbpl, maxline);

  free (slice);
}

/*******************************************
 * Scan image (6000SP)       XXX           *
 *******************************************/
static void scan_image_6000sp
  (
    int param			/* I: scanner type */
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char testunit[] =
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char req_sense[] =
  {0x03, 0x00, 0x00, 0x00, 0x04, 0x00};
  unsigned char area[] =
  {0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0, 0, 0, 0, 0, 0, 0};
  unsigned char mode[] =
  {0x15, 0x00, 0x00, 0x00, 0x0d, 0x00, 0x83, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char lookup[] =
  {0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char start[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char stop[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char status[] =
  {0x0f, 0x00, 0x00, 0x00, 0x06, 0x00};
  unsigned char backtrack[] =
  {0x10, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char readline[] =
  {0x08, 0x00, 0x00, 0x00, 0x01, 0x00};
#ifndef SG_BIG_BUFF
  unsigned char image[4096];
#else
  unsigned char image[SG_BIG_BUFF];
#endif
  unsigned char state[6];
  int i, line, maxline, maxscanline, bpl, maxbpl;
  int maxcol;
  int singlemaxbpl;
  int x, y;
  unsigned char *slice = NULL;
  int ll;
  int allowed_res[3] =
  {600, 800, 1200};
  int blind, coloffset, allocsize;

#define MAXRETRY 8
#define BIGRETRY 3

  TRACE (1, "scan image\n");

  if (backtracking < 0)
    backtracking = 0;

  /* MFS-12000 seems to rescan blind area after backtracking */
  blind = backtracking ? ((color == 1) ? resolution / 33 : resolution / 90) : 0;
  maxcol = (color == 1 ? 3 : 1);
  coloffset = (color == 1) ? resolution / 72 : 0;
  singlemaxbpl = resolution * x_size;
  maxbpl = singlemaxbpl * maxcol;
  maxline = resolution * y_size;
  maxscanline = maxline + 2 * coloffset;

  if (resolution > allowed_res[param] || resolution < 3)
    ERROR ("need resolution in the range of 3..%d\n", allowed_res[param]);

  if (y_size + y_pos > 14.0)
    ERROR ("illegal size/length: would run over border by %2.2f\"\n", y_size + y_pos - 14.0);

  if (x_size + x_pos > 9.0)
    ERROR ("illegal size/width: would run over border by %2.2f\"\n", x_size + x_pos - 9.0);

  /* MFS-12000 seems to rescan blind area after backtracking */
  blind = backtracking ? ((color == 1) ? resolution / 33 : resolution / 90) : 0;

  allocsize = maxbpl * maxline;
  if ((slice = malloc (allocsize)) == NULL)
    ERROR ("cannot malloc %d bytes\n", allocsize);
  memset (slice, 0, allocsize);

#if 0
  if (maxbpl >= 4095)
    ERROR ("line overflow (%d, allowed 4095) - use lower resolution or x-size\n", maxbpl);
#endif

  {
    int busy, lib, retry, bigretry;
    unsigned char mask;
    int loop;
    int singlebpl;

    bigretry = 0;
    retry = 0;

  scan_retry:
    scsi_cmd (inquiry, 6, image, 256);

    scsi_cmd (testunit, 6, NULL, 0);


    area[6] = maxcol;
    area[7] = ((unsigned int) (x_pos * 16)) & 0xff;
    area[8] = (((unsigned int) (x_pos * 16)) & 0xff00) >> 8;
    area[9] = ((unsigned int) (y_pos * 16)) & 0xff;
    area[10] = (((unsigned int) (y_pos * 16)) & 0xff00) >> 8;

    area[11] = ((unsigned int) ((x_pos + x_size) * 16)) & 0xff;
    area[12] = (((unsigned int) ((x_pos + x_size) * 16)) & 0xff00) >> 8;
    area[13] = ((unsigned int) ((y_pos + y_size + coloffset * 2) * 16)) & 0xff;
    area[14] = (((unsigned int) ((y_pos + y_size + coloffset * 2) * 16)) & 0xff00) >> 8;

    scsi_cmd (area, 15, NULL, 0);

    scsi_cmd (req_sense, 6, image, 4);
    if (image[0] != 0)
      ERROR ("sense Error: %dd 0x%x\n", image[0], image[0]);

    for (loop = 0; loop < maxcol; loop++)
      {
	mask = (color == 1) ? (loop == 0 ? 0x20 : (loop == 1 ? 0x40 : 0x60)) : 0;

	mode[6] = 0x83 | mask;
	mode[8] = brightness & 0xff;
	mode[9] = contrast & 0xff;
	mode[10] = grain & 0xff;
	mode[11] = speed & 0xff;
	mode[14] = ((unsigned int) ((x_size + x_pos) * 8)) & 0xff;
	mode[15] = (((unsigned int) ((x_size + x_pos) * 8)) & 0xff00) >> 8;
	mode[17] = resolution % 256;
	mode[18] = resolution / 256;

	scsi_cmd (mode, 19, NULL, 0);
      }

    scsi_cmd (req_sense, 6, image, 4);
    if (image[0] != 0)
      ERROR ("sense Error: %dd 0x%x\n", image[0], image[0]);

    scsi_cmd (lookup, 10, NULL, 0);

    mask = (color == 1) ? 0x20 : 0;

    if (color == 0 || color == 1)
      start[4] = 0xc1 | mask;
    else
      start[4] = 0x81 | mask;
    scsi_cmd (start, 6, NULL, 0);
    scan_running = 1;

    line = 0;

    retry = 0;

    do
      {
	scsi_cmd (status, 6, state, 6);
	busy = state[0];
	bpl = state[1] + state[2] * 256;
	lib = state[3] + state[4] * 256 + state[5] * 65536;

	TRACE (2, "rdy %d bpl(%d) %4d lib(%d) %8d\n", !busy, maxbpl, bpl, maxscanline, lib);

	retry++;
	if (busy)
	  sleep (retry);

	if (retry > MAXRETRY)
	  break;
      }
    while (busy);

    singlebpl = bpl / maxcol;

#if 0
    if (bpl > 4095 || busy)
#else
    if (busy)
#endif
      {
	if (bigretry < BIGRETRY)
	  {
	    bigretry++;
	    scsi_cmd (stop, 6, NULL, 0);
	    scan_running = 0;

	    sleep (2);

	    TRACE (1, "bigretry %d\n", bigretry);

	    goto scan_retry;
	  }
      }

    backtrack[4] = (backtracking ? 0x82 : 0x80);
    scsi_cmd (backtrack, 6, NULL, 0);

    if (retry <= MAXRETRY)
      do
	{
	  int col, putline, minx;

#ifdef SG_BIG_BUFF
	  ll = (mem_size - sizeof (struct sg_header)) / bpl;
	  if (ll > maxscanline + blind - line)
	    ll = maxscanline + blind - line;
	  if (ll < 1)
	    ll = 1;
#else
	  ll = 1;
#endif

	  if (ll < 2)
	    ll = 2;		/* MFS-12000SP hangs if ll==1 */

	  if (ll <= blind)
	    ERROR ("blind area (%d) larger than lines per scan(%d): either decrease resolution or x_size\n", blind, ll);

	  readline[2] = ((ll & 0xff0000) >> 16) & 0xff;
	  readline[3] = ((ll & 0x00ff00) >> 8) & 0xff;
	  readline[4] = ((ll & 0x0000ff)) & 0xff;

	  TRACE (1, "ll=%d, blind=%d, bpl=%d, maxbpl=%d, line=%d, maxline=%d, maxscanline=%d\n", ll, blind, bpl, maxbpl, line, maxline, maxscanline);
	  scsi_cmd (readline, 6, image, bpl * ll);

	  minx = (singlebpl < singlemaxbpl) ? singlebpl : singlemaxbpl;
	  for (y = blind; y < ll; y++, line++)
	    for (putline = line, col = 0; col < maxcol; col++, putline -= coloffset)
	      if (putline < maxline)
		for (x = 0; x < minx; x++)
		  {
		    int index = x + (putline + col * maxline) * singlemaxbpl;
		    if (index >= 0)
		      slice[index] = image[x + y * bpl + col * singlebpl];
		  }
	  TRACE (2, "Line progress ll:%d line:%d\n", ll, line);

	  scsi_cmd (status, 6, state, 6);
	  busy = state[0];
	  bpl = state[1] + state[2] * 256;
	  lib = state[3] + state[4] * 256 + state[5] * 65536;

	  TRACE (2, "rdy %d bpl(%d) %4d lib(%ul) %8ul\n", !busy, maxbpl, bpl, lib, maxscanline);
	}
      while (line < maxscanline);

    TRACE (2, "lines processed %d, bpl processed %d, maxbpl %d, maxscanline %d\n", line, bpl, maxbpl, maxscanline);

    scsi_cmd (lookup, 10, NULL, 0);

    scsi_cmd (stop, 6, NULL, 0);
    scan_running = 0;

    sleep ((int) (y_pos + y_size) + 3);

    for (i = 1; i <= MAXRETRY; i++)
      {
	scsi_cmd (status, 6, state, 6);
	if (!state[0])
	  break;
	sleep (i);
      }

    scsi_cmd (stop, 6, NULL, 0);
    scan_running = 0;

    if (retry > MAXRETRY)
      ERROR ("scanner timed out\n");

#if 0
    if (bpl > 4095)
      ERROR ("input line too long (%d, allowed 4095) - use lower resolution or x-size\n", bpl);
#endif
  }

  if (outputformat == 0)
    write_image (slice, singlemaxbpl, maxline);
  else
    write_print_image (slice, singlemaxbpl, maxline);

  free (slice);
}

/*******************************************
 * Main procedure                          *
 *******************************************/
void main
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  signal (SIGHUP, signal_handler);
  signal (SIGINT, signal_handler);
  signal (SIGQUIT, signal_handler);
  signal (SIGPIPE, signal_handler);
  signal (SIGALRM, signal_handler);
  signal (SIGTERM, signal_handler);
  signal (SIGXCPU, signal_handler);
  signal (SIGXFSZ, signal_handler);
  signal (SIGVTALRM, signal_handler);
  signal (SIGPROF, signal_handler);
  signal (SIGUSR1, signal_handler);
  signal (SIGUSR2, signal_handler);

  parse_commandline (argc, argv);

  open_scanner ();

  scanner_type[scanner_index].scan (scanner_type[scanner_index].param);

  close_scanner ();
}

/*
   End of file
 */
